from singer.logger import get_logger

from .abstract import IncrementalTableStream

LOGGER = get_logger()


class PerformanceReportCustom(IncrementalTableStream):
    """Class Representing the `performance_report_custom` Stream."""

    tap_stream_id = "performance_report_custom"
    key_properties = ["site_url", "search_type", "date", "dimensions_hash_key"]
    valid_replication_keys = ("date",)

    body_params = {"aggregationType": "auto"}
    dimension_list = ["date", "country", "device", "page", "query"]


class PerformanceReportDate(IncrementalTableStream):
    """Class Representing the `performance_report_date` Stream."""

    tap_stream_id = "performance_report_date"
    key_properties = ["site_url", "search_type", "date"]
    valid_replication_keys = ("date",)

    body_params = {"aggregationType": "byProperty", "dimensions": ["date"]}


class PerformanceReportCountry(IncrementalTableStream):
    """Class Representing the `performance_report_country` Stream."""

    tap_stream_id = "performance_report_country"
    key_properties = ["site_url", "search_type", "date", "country"]
    valid_replication_keys = ("date",)

    body_params = {"aggregationType": "byProperty", "dimensions": ["date", "country"]}


class PerformanceReportDevices(IncrementalTableStream):
    """Class Representing the `performance_report_device` Stream."""

    tap_stream_id = "performance_report_device"
    key_properties = ["site_url", "search_type", "date", "device"]
    # Excluding discover sub_type since Requests for Discover cannot be grouped by device.
    # this also overwrites the sub_types attribute declared for class IncrementalTableStream
    # in abstract.py
    sub_types = ["googleNews", "image", "news", "video", "web"]
    valid_replication_keys = ("date",)

    body_params = {"aggregationType": "byProperty", "dimensions": ["date", "device"]}


class PerformanceReportPage(IncrementalTableStream):
    """Class Representing the `performance_report_page` Stream."""

    tap_stream_id = "performance_report_page"
    key_properties = ["site_url", "search_type", "date", "page"]
    valid_replication_keys = ("date", "page")

    body_params = {"aggregationType": "byPage", "dimensions": ["date", "page"]}


class PerformanceReportQuery(IncrementalTableStream):
    """Class Representing the `performance_report_query` Stream."""

    tap_stream_id = "performance_report_query"
    key_properties = ["site_url", "search_type", "date", "query"]
    # Excluding discover and googleNews since query seems to be an invalid argument while
    # grouping data for discover and googleNews
    # this also overwrites the sub_types attribute declared for class IncrementalTableStream
    # in abstract.py
    sub_types = ["image", "news", "video", "web"]
    valid_replication_keys = ("date",)

    body_params = {"aggregationType": "byProperty", "dimensions": ["date", "query"]}



from typing import Dict, Iterator

from singer.logger import get_logger

from tap_google_search_console.helpers import encode_and_format_url, transform_json

from .abstract import FullTableStream

LOGGER = get_logger()


class Sitemaps(FullTableStream):
    """Class Representing the `Sitemaps` Stream."""

    tap_stream_id = "sitemaps"
    key_properties = ["site_url", "path", "last_submitted"]

    data_key = "sitemap"
    path = "sites/{}/sitemaps"

    def get_records(self) -> Iterator[Dict]:
        """Performs API calls to extract data for each site."""
        LOGGER.info("get records called from %s", self.__class__)
        transformed_data = []
        for site in self.get_site_url():
            if site[:9] == "sc-domain":
                LOGGER.info(f"Skipping Site: {site}")
                LOGGER.info("Sitemaps API does not support domain property urls at this time.")
                continue
            path = encode_and_format_url(site, self.path)
            data = self.client.get(path)
            if not data:
                return transformed_data
            transformed_data.extend(
                iter(transform_json(data, self.tap_stream_id, site=site, path=self.data_key).get(self.data_key, []))
            )
        yield from transformed_data



from .performance_reports import (
    PerformanceReportCountry,
    PerformanceReportCustom,
    PerformanceReportDate,
    PerformanceReportDevices,
    PerformanceReportPage,
    PerformanceReportQuery,
)
from .sitemaps import Sitemaps
from .sites import Sites

STREAMS = {
    Sites.tap_stream_id: Sites,
    Sitemaps.tap_stream_id: Sitemaps,
    PerformanceReportCustom.tap_stream_id: PerformanceReportCustom,
    PerformanceReportDate.tap_stream_id: PerformanceReportDate,
    PerformanceReportCountry.tap_stream_id: PerformanceReportCountry,
    PerformanceReportDevices.tap_stream_id: PerformanceReportDevices,
    PerformanceReportPage.tap_stream_id: PerformanceReportPage,
    PerformanceReportQuery.tap_stream_id: PerformanceReportQuery,
}



import json
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Dict, List, Tuple, Union

from singer import (
    Transformer,
    metadata,
    metrics,
    should_sync_field,
    utils,
    write_record,
    write_state,
)
from singer.logger import get_logger
from singer.metadata import get_standard_metadata

from tap_google_search_console.helpers import encode_and_format_url, transform_json

LOGGER = get_logger()


class BaseStream(ABC):
    """Base class representing generic stream methods and meta-attributes."""

    @property
    @abstractmethod
    def replication_method(self) -> str:
        """Defines the sync mode of a stream."""

    @property
    @abstractmethod
    def forced_replication_method(self) -> str:
        """Defines the sync mode of a stream."""

    @property
    @abstractmethod
    def replication_key(self) -> str:
        """Defines the replication key for incremental sync mode of a
        stream."""

    @property
    @abstractmethod
    def valid_replication_keys(self) -> Tuple[str, str]:
        """Defines the replication key for incremental sync mode of a
        stream."""

    @property
    @abstractmethod
    def key_properties(self) -> Tuple[str, str]:
        """List of key properties for stream."""

    @property
    @abstractmethod
    def tap_stream_id(self) -> str:
        """The unique identifier for the stream.

        This is allowed to be different from the name of the stream in
        order to allow for sources that have duplicate stream names.
        """

    @abstractmethod
    def sync(self, state: Dict, schema: Dict, stream_metadata: Dict):
        """Performs Sync."""

    def __init__(self, client=None, config=None) -> None:
        self.client = client
        self.config = config

    @classmethod
    def get_metadata(cls, schema) -> Dict[str, str]:
        """Returns a `dict` for generating stream metadata."""
        return get_standard_metadata(**{
            "schema": schema,
            "key_properties": cls.key_properties,
            "valid_replication_keys": cls.valid_replication_keys,
            "replication_method": cls.replication_method or cls.forced_replication_method,
        }
                                         )

    def get_site_url(self):
        return self.config.get("site_urls", "").replace(" ", "").split(",")


class IncrementalTableStream(BaseStream, ABC):
    """Base Class for Incremental Stream."""

    replication_method = "INCREMENTAL"
    forced_replication_method = "INCREMENTAL"
    replication_key = "date"
    pagination = "body"
    sub_types = ["discover", "googleNews", "image", "news", "video", "web"]
    row_limit = 10000
    path = "sites/{}/searchAnalytics/query"
    data_key = "rows"
    now_dt_tm = utils.now()
    dimension_list = []
    body_params = {}

    # declaring this variable to keep track of number of
    # records processed per stream, per site, per sub_type
    records_extracted = 0

    @staticmethod
    def get_bookmark(state: Dict, stream: str, site: str, sub_type: str, default: str) -> str:
        """Fetches the bookmark from the state file for a given stream, site,
        sub_type."""

        if (state is None) or ("bookmarks" not in state):
            return default
        return state.get("bookmarks", {}).get(stream, {}).get(site, {}).get(sub_type, default)

    @property
    def get_attribution_days(self) -> Union[str, int]:
        """Sets this value to 4days since there is data delay of 2-3 days from GSC"""
        return int(self.config.get("ATTRIBUTION_DAYS") or 4)

    @property
    def get_date_window_size(self) -> Union[str, int]:
        return int(self.config.get("DATE_WINDOW_SIZE") or 30)

    def write_bookmark(self, state: Dict, site: str, sub_type: str, value: str) -> None:
        """Writes bookmark to state file for a given stream, site, sub_type."""
        if "bookmarks" not in state:
            state["bookmarks"] = {}
        if self.tap_stream_id not in state["bookmarks"]:
            state["bookmarks"][self.tap_stream_id] = {}
        if site not in state["bookmarks"][self.tap_stream_id]:
            state["bookmarks"][self.tap_stream_id][site] = {}
        state["bookmarks"][self.tap_stream_id][site][sub_type] = value
        LOGGER.info(f"Write state for Stream: {self.tap_stream_id}, Site: {site}, Type: {sub_type}, value: {value}")
        write_state(state)

    def set_start_and_end_times(self, state: Dict, stream: str, sub_type: str, site: str) -> Tuple[datetime, datetime]:
        """Method to set start and end times."""
        # get the bookmark from state file
        report_bookmark = self.get_bookmark(state, stream, site, sub_type, self.config.get("start_date"))
        reports_dt_tm = utils.strptime_to_utc(report_bookmark)

        delta_days = (self.now_dt_tm - reports_dt_tm).days

        if delta_days < self.get_attribution_days:
            start_dt_tm = self.now_dt_tm - timedelta(days=self.get_attribution_days)
        else:
            start_dt_tm = reports_dt_tm

        end_dt_tm = start_dt_tm + timedelta(days=self.get_date_window_size)
        if end_dt_tm > self.now_dt_tm:
            end_dt_tm = self.now_dt_tm

        return start_dt_tm, end_dt_tm

    def set_dimensions_in_payload(self, stream_metadata: Dict) -> List:
        """Set only the selected (field selection) dimensions in API
        payload."""
        selected_dimensions = []
        for dimension in self.dimension_list:
            if should_sync_field(
                metadata.get(stream_metadata, ("properties", dimension), "inclusion"),
                metadata.get(stream_metadata, ("properties", dimension), "selected"),
            ):
                selected_dimensions.append(dimension)
        return selected_dimensions

    def make_payload(self, sub_type: str, start_date: str, end_date: str, stream_metadata: Dict) -> Dict:
        """Creates payload for POST API Call."""
        if self.tap_stream_id == "performance_report_custom":
            self.body_params["dimensions"] = self.set_dimensions_in_payload(stream_metadata)
            # Remove discover dimension from dimension_list if sub_type is discover
            # Requests for Discover cannot be grouped by device
            if sub_type == "discover" and "device" in self.body_params["dimensions"]:
                LOGGER.info(f"Removing the device dimension/field since it is incompatible with"
                            f" {sub_type} sub_type for custom report")
                self.body_params["dimensions"].remove("device")
        if sub_type in {"discover", "googleNews"}:
            self.body_params["aggregationType"] = "auto"
            # Remove query from dimension list if the sub_type is either discover or googleNews
            # query seems to be an invalid argument while grouping data for discover and googleNews
            if self.tap_stream_id == "performance_report_custom" and \
                    "query" in self.body_params["dimensions"]:
                LOGGER.info(f"Removing the query dimension/field since it is incompatible with"
                            f" {sub_type} sub_type for custom report")
                self.body_params["dimensions"].remove("query")

        return {"type": sub_type, "startDate": start_date, "endDate": end_date, **self.body_params}

    def validate_keys_in_data(self, extracted_data: List) -> None:
        """Validates the data by checking the primary keys in extracted
        data."""
        for record in extracted_data:
            for key in self.key_properties:
                if not record.get(key):
                    primary_keys_only = {id_field: record.get(id_field) for id_field in self.key_properties}
                    raise ValueError(f"Missing key {key} in record with primary keys {primary_keys_only}")

    def modify_start_end_dt_tm(self, end_dt_tm: datetime) -> Tuple[datetime, datetime]:
        """Sets start_date_time of a new window to end_date_time of old window
        Sets end_date_time of new window to 30 days(date_window_size) ahead
        since start_date_time."""
        start_dt_tm = end_dt_tm
        end_dt_tm = start_dt_tm + timedelta(days=self.get_date_window_size)
        if end_dt_tm > self.now_dt_tm:
            end_dt_tm = self.now_dt_tm
        return start_dt_tm, end_dt_tm

    def process_records(
        self,
        schema: Dict,
        stream_metadata: Dict,
        records: List,
        time_extracted: datetime,
        max_bookmark_value=None,
        last_datetime=None,
    ) -> str:
        """Filters out the unselected fields by the user Picks the latest
        bookmark value from extracted data Writes the records to stdout."""

        with metrics.record_counter(self.tap_stream_id) as counter:
            for record in records:
                # Transform record for Singer.io
                with Transformer() as transformer:
                    transformed_record = transformer.transform(record, schema, stream_metadata)

                    # Reset max_bookmark_value to new value if higher
                    if self.replication_key in transformed_record:
                        bookmark_date = transformed_record.get(self.replication_key)
                        bookmark_dt_tm = utils.strptime_to_utc(bookmark_date)
                        last_dt_tm = utils.strptime_to_utc(last_datetime)

                        if not max_bookmark_value:
                            max_bookmark_value = last_datetime

                        max_bookmark_dt_tm = utils.strptime_to_utc(max_bookmark_value)

                        if bookmark_dt_tm > max_bookmark_dt_tm:
                            max_bookmark_value = utils.strftime(bookmark_dt_tm)

                        # Keep only records whose bookmark is after the last_datetime
                        if bookmark_dt_tm >= last_dt_tm:
                            write_record(self.tap_stream_id, transformed_record, time_extracted=time_extracted)
                            counter.increment()
                    else:
                        write_record(self.tap_stream_id, transformed_record, time_extracted=time_extracted)
                        counter.increment()

            LOGGER.info(f"Stream: {self.tap_stream_id}, Processed {counter.value} records")
            self.records_extracted += counter.value
            return max_bookmark_value

    def get_records_for_sub_type(
        self, site_url: str, sub_type: str, state: Dict, schema: Dict, stream_metadata: Dict
    ) -> None:
        """Sync the data for a given sub_type, stream, site Gets the bookmark
        value or start date value, extracts data for date window size of 30
        days."""
        start_dt_tm, end_dt_tm = self.set_start_and_end_times(state, self.tap_stream_id, sub_type, site_url)
        LOGGER.info(f"bookmark value or start date for {self.tap_stream_id} {site_url} {sub_type}: {start_dt_tm}")
        site_path = encode_and_format_url(site_url, self.path)
        while start_dt_tm < end_dt_tm:
            offset, row_limit, batch_count = 0, self.row_limit, self.row_limit
            last_datetime = self.get_bookmark(
                state, self.tap_stream_id, site_url, sub_type, self.config.get("start_date")
            )
            bookmark_value = last_datetime
            start_str, end_str = utils.strftime(start_dt_tm)[:10], utils.strftime(end_dt_tm)[:10]

            LOGGER.info(
                f"Running sync for {site_url}, {self.tap_stream_id}, {sub_type} between date window "
                f"{start_str} {end_str}"
            )
            payload = self.make_payload(sub_type, start_str, end_str, stream_metadata)
            while row_limit == batch_count:
                body = {"startRow": offset, "rowLimit": row_limit, **payload}
                time_extracted = utils.now()
                LOGGER.info(f"body = {body}")
                data = self.client.post(site_path, endpoint=self.tap_stream_id, data=json.dumps(body))
                if not data:
                    self.write_bookmark(state, site_url, sub_type, bookmark_value)
                    LOGGER.info(f"There are no raw data records for date window {start_dt_tm} to {end_dt_tm}, "
                                f" from offset value {offset}")
                    start_dt_tm, end_dt_tm = self.modify_start_end_dt_tm(end_dt_tm)
                transformed_data = []
                if self.data_key in data:
                    transformed_data = transform_json(
                        data,
                        self.tap_stream_id,
                        self.data_key,
                        site_url,
                        sub_type,
                        dimensions_list=payload.get("dimensions", []),
                    )[self.data_key]

                if not transformed_data:
                    self.write_bookmark(state, site_url, sub_type, bookmark_value)
                    start_dt_tm, end_dt_tm = self.modify_start_end_dt_tm(end_dt_tm)

                self.validate_keys_in_data(transformed_data)
                LOGGER.info(f"Total synced records for {sub_type} {self.tap_stream_id}: {len(transformed_data)}")
                batch_count = len(transformed_data)
                bookmark_value = self.process_records(
                    schema,
                    stream_metadata,
                    transformed_data,
                    time_extracted,
                    bookmark_value,
                    last_datetime=last_datetime,
                )
                self.write_bookmark(state, site_url, sub_type, bookmark_value)
                offset = offset + row_limit

            start_dt_tm, end_dt_tm = self.modify_start_end_dt_tm(end_dt_tm)

    def get_records_for_site(self, site_url: str, state: Dict, schema: Dict, stream_metadata: Dict) -> None:
        """Starts Syncing data for each sub_type for a given site Logs the
        total number of extracted records."""
        for sub_type in self.sub_types:
            LOGGER.info(f"Starting Sync for Stream {self.tap_stream_id}, Site {site_url}, Type {sub_type}")
            self.records_extracted = 0
            self.get_records_for_sub_type(site_url, sub_type, state, schema, stream_metadata)
            LOGGER.info(
                f"Total records extracted for Stream: {self.tap_stream_id}, Site: {site_url}, Type: {sub_type}:"
                f" {self.records_extracted}"
            )
            LOGGER.info(f"Finished Sync for Stream {self.tap_stream_id}, Site {site_url}, Type {sub_type}")

    def get_records(self, state: Dict, schema: Dict, stream_metadata: Dict) -> None:
        """starts extracting data for each site_url configured by the user."""
        for site in self.get_site_url():
            LOGGER.info(f"Starting Sync for Stream {self.tap_stream_id}, Site {site}")
            self.get_records_for_site(site, state, schema, stream_metadata)
            LOGGER.info(f"Finished Sync for Stream {self.tap_stream_id}, Site {site}")

    def sync(self, state: Dict, schema: Dict, stream_metadata: Dict) -> None:
        """Starts Sync."""
        LOGGER.info(f"Starting Sync for Stream {self.tap_stream_id}")
        self.get_records(state, schema, stream_metadata)
        LOGGER.info(f"Finished Sync for Stream {self.tap_stream_id}")


class FullTableStream(BaseStream, ABC):
    """Base Class for Incremental Stream."""

    replication_method = "FULL_TABLE"
    forced_replication_method = "FULL_TABLE"
    api_method = "GET"
    valid_replication_keys = None
    replication_key = None

    @abstractmethod
    def get_records(self):
        """Extracts Records."""

    def sync(self, state: Dict, schema: Dict, stream_metadata: Dict):
        LOGGER.info("sync called from %s", self.__class__)
        with metrics.record_counter(self.tap_stream_id) as counter:
            time_extracted = utils.now()
            for record in self.get_records():
                with Transformer() as transformer:
                    transformed_record = transformer.transform(record, schema, stream_metadata)
                    write_record(self.tap_stream_id, transformed_record, time_extracted=time_extracted)
                    counter.increment()



import json
from datetime import datetime, timedelta, timezone
from typing import Any
from urllib.parse import quote

import backoff
import requests
import singer
from requests.exceptions import ConnectionError, Timeout
from singer import metrics, utils

from .exceptions import (
    GoogleQuotaExceededError,
    GoogleRateLimitExceeded,
    Server5xxError,
    raise_for_error,
)

BASE_URL = "https://www.googleapis.com/webmasters/v3"
GOOGLE_TOKEN_URI = "https://oauth2.googleapis.com/token"
LOGGER = singer.get_logger()

# set default timeout of 300 seconds
REQUEST_TIMEOUT = 300


class GoogleClient:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        refresh_token: str,
        site_urls: str,
        user_agent=None,
        timeout=REQUEST_TIMEOUT,
    ):

        self.__client_id, self.__client_secret, self.__refresh_token = (client_id, client_secret, refresh_token)
        self.__site_urls, self.__user_agent = site_urls, user_agent
        self.__access_token, self.__expires, self.base_url = None, None, None
        self.__session = requests.Session()

        try:
            self.request_timeout = REQUEST_TIMEOUT if timeout in (None, 0, "0", "0.0") else float(timeout)
        except ValueError:
            self.request_timeout = REQUEST_TIMEOUT

    def check_sites_access(self) -> None:
        """Perform access check for each site url provided."""
        body = json.dumps({"startDate": "2021-04-01", "endDate": "2021-05-01"})
        for site_url in self.__site_urls.replace(" ", "").split(","):
            self.post(f"sites/{quote(site_url, safe='')}/searchAnalytics/query", data=body)

    @backoff.on_exception(backoff.expo, (Server5xxError, ConnectionError, Timeout), max_tries=5, factor=2)
    def __enter__(self):
        self.get_access_token()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.__session.close()

    def get_access_token(self) -> None:
        """Performs authentication and the access token if expired."""

        if self.__access_token and self.__expires > datetime.now(timezone.utc):
            return
        headers = {"User-Agent": self.__user_agent or ""}
        response = self.__session.post(
            url=GOOGLE_TOKEN_URI,
            headers=headers,
            data={
                "grant_type": "refresh_token",
                "client_id": self.__client_id,
                "client_secret": self.__client_secret,
                "refresh_token": self.__refresh_token,
            },
            timeout=self.request_timeout,
        )

        if response.status_code != 200:
            raise_for_error(response)

        data = response.json()
        self.__access_token = data["access_token"]
        self.__expires = utils.now() + timedelta(seconds=data["expires_in"])

        LOGGER.info("Authorized, token expires = %s", self.__expires)

    # Backoff for 15 minutes in case of Quota Exceeded error
    @backoff.on_exception(backoff.constant, GoogleQuotaExceededError, max_tries=2, interval=900, jitter=None)
    # backoff for 5 times, with 10 seconds consistent interval
    @backoff.on_exception(backoff.constant, Timeout, max_tries=5, interval=10, jitter=None)
    @backoff.on_exception(
        backoff.expo, (Server5xxError, ConnectionError, GoogleRateLimitExceeded), max_tries=7, factor=3
    )
    @utils.ratelimit(1200, 60)
    def request(self, method: str, path: str = None, url: str = None, **kwargs) -> Any:
        """Wrapper method around request.sessions get/post method using the
        session object of the GoogleClient Object."""

        # TODO: Consolidate multiple backoff decorators
        self.get_access_token()
        url = url or f"{self.base_url or BASE_URL}/{path}"

        endpoint, kwargs["headers"] = kwargs.get("endpoint", None), kwargs.get("headers", {})
        kwargs.pop("endpoint", None)

        kwargs["headers"]["Authorization"] = f"Bearer {self.__access_token}"
        if self.__user_agent:
            kwargs["headers"]["User-Agent"] = self.__user_agent
        if method == "POST":
            kwargs["headers"]["Content-Type"] = "application/json"

        with metrics.http_request_timer(endpoint) as timer:
            response = self.__session.request(method, url, timeout=self.request_timeout, **kwargs)
            timer.tags[metrics.Tag.http_status_code] = response.status_code

        if response.status_code != 200:
            raise_for_error(response)

        return response.json()

    def get(self, path: str, **kwargs) -> Any:
        """wrapper for get method."""
        return self.request("GET", path=path, **kwargs)

    def post(self, path: str, **kwargs) -> Any:
        """wrapper for post method."""
        return self.request("POST", path=path, **kwargs)



from typing import Dict, Iterator

from singer.logger import get_logger

from tap_google_search_console.helpers import encode_and_format_url, transform_json

from .abstract import FullTableStream

LOGGER = get_logger()


class Sites(FullTableStream):
    """Class Representing the `Sites` Stream."""

    tap_stream_id = "sites"
    key_properties = [
        "site_url",
    ]

    data_key = "site_entry"
    path = "sites/{}"

    def get_records(self) -> Iterator[Dict]:
        """Performs API calls to extract data for each site."""
        records = []
        for site in self.get_site_url():
            path = encode_and_format_url(site, self.path)
            data = self.client.get(path)
            records.append(data)
        data = {self.data_key: records}
        # transforms data by converting camelCase fields to snake_case fields
        transformed_records = transform_json(data, self.tap_stream_id)
        yield from transformed_records.get(self.data_key, [])



import json
from typing import Dict, Tuple

from singer.catalog import Catalog

from tap_google_search_console.client import GoogleClient
from tap_google_search_console.helpers import get_abs_path
from tap_google_search_console.streams import STREAMS


def get_schemas() -> Tuple[Dict, Dict]:
    """Builds the singer schema and metadata dictionaries."""
    streams, stream_metadata = {}, {}

    for stream_name, stream in STREAMS.items():
        schema_path = get_abs_path(f"schemas/{stream_name}.json")

        with open(schema_path, encoding="utf-8") as file:
            schema = json.load(file)

        streams[stream_name], stream_metadata[stream_name] = schema, stream.get_metadata(schema)

    return streams, stream_metadata


def discover(client: GoogleClient):
    """Performs permission check and starts discover."""
    client.check_sites_access()
    schemas, schema_metadata = get_schemas()
    streams = []

    for schema_name, schema in schemas.items():
        schema_meta = schema_metadata[schema_name]
        streams.append({"stream": schema_name, "tap_stream_id": schema_name, "schema": schema, "metadata": schema_meta})
    return Catalog.from_dict({"streams": streams})



from typing import Dict

import singer
from singer import Catalog, metadata

from .client import GoogleClient as Client
from .streams import STREAMS

LOGGER = singer.get_logger()


def sync(client: Client, config: Dict, state: Dict, catalog: Catalog):
    """Sync data from tap source"""

    for stream in catalog.get_selected_streams(state):
        tap_stream_id = stream.tap_stream_id
        stream_obj = STREAMS[tap_stream_id](client, config)
        stream_schema = stream.schema.to_dict()
        stream_metadata = metadata.to_map(stream.metadata)

        LOGGER.info("Starting sync for stream: %s", tap_stream_id)

        state = singer.set_currently_syncing(state, tap_stream_id)
        singer.write_state(state)

        singer.write_schema(tap_stream_id, stream_schema, stream_obj.key_properties, stream.replication_key)

        stream_obj.sync(state, stream_schema, stream_metadata)

    state = singer.set_currently_syncing(state, None)
    singer.write_state(state)



import json


class GoogleError(Exception):
    pass


class Server5xxError(GoogleError):
    pass


class GoogleBadRequestError(GoogleError):
    pass


class GoogleUnauthorizedError(GoogleError):
    pass


class GooglePaymentRequiredError(GoogleError):
    pass


class GoogleNotFoundError(GoogleError):
    pass


class GoogleMethodNotAllowedError(GoogleError):
    pass


class GoogleConflictError(GoogleError):
    pass


class GoogleGoneError(GoogleError):
    pass


class GooglePreconditionFailedError(GoogleError):
    pass


class GoogleRequestEntityTooLargeError(GoogleError):
    pass


class GoogleRequestedRangeNotSatisfiableError(GoogleError):
    pass


class GoogleExpectationFailedError(GoogleError):
    pass


class GoogleForbiddenError(GoogleError):
    pass


class GoogleUnprocessableEntityError(GoogleError):
    pass


class GooglePreconditionRequiredError(GoogleError):
    pass


class GoogleRateLimitExceeded(GoogleError):
    pass


class GoogleInternalServiceError(Server5xxError):
    pass


class GoogleNotImplementedError(Server5xxError):
    pass


class GoogleServiceUnavailable(Server5xxError):
    pass


class GoogleQuotaExceededError(GoogleError):
    pass


class GoogleInvalidGrant(GoogleError):
    pass


# Error Codes: https://developers.google.com/webmaster-tools/search-console-api-original/v3/errors
ERROR_CODE_EXCEPTION_MAPPING = {
    400: {"raise_exception": GoogleBadRequestError, "message": "The request is missing or has bad parameters."},
    401: {"raise_exception": GoogleUnauthorizedError, "message": "Invalid authorization credentials."},
    402: {
        "raise_exception": GooglePaymentRequiredError,
        "message": "The requested operation requires more resources than the quota allows. Payment is required to "
        "complete the operation.",
    },
    403: {"raise_exception": GoogleForbiddenError, "message": "Invalid authorization credentials or permissions."},
    404: {"raise_exception": GoogleNotFoundError, "message": "The requested resource does not exist."},
    405: {
        "raise_exception": GoogleMethodNotAllowedError,
        "message": "The HTTP method associated with the request is not supported.",
    },
    409: {
        "raise_exception": GoogleConflictError,
        "message": "The API request cannot be completed because the requested operation would conflict with an "
        "existing item.",
    },
    410: {"raise_exception": GoogleGoneError, "message": "The requested resource is permanently unavailable."},
    412: {
        "raise_exception": GooglePreconditionFailedError,
        "message": "The condition set in the request's If-Match or If-None-Match HTTP request header was not met.",
    },
    413: {"raise_exception": GoogleRequestEntityTooLargeError, "message": "The request is too large."},
    416: {
        "raise_exception": GoogleRequestedRangeNotSatisfiableError,
        "message": "The request specified a range that cannot be satisfied.",
    },
    417: {
        "raise_exception": GoogleExpectationFailedError,
        "message": "A client expectation cannot be met by the server.",
    },
    422: {
        "raise_exception": GoogleUnprocessableEntityError,
        "message": "The request was not able to process right now.",
    },
    428: {
        "raise_exception": GooglePreconditionRequiredError,
        "message": "The request requires a precondition If-Match or If-None-Match which is not provided.",
    },
    429: {"raise_exception": GoogleRateLimitExceeded, "message": "Rate limit has been exceeded."},
    500: {"raise_exception": GoogleInternalServiceError, "message": "The request failed due to an internal error."},
    501: {"raise_exception": GoogleNotImplementedError, "message": "Functionality does not exist."},
    503: {"raise_exception": GoogleServiceUnavailable, "message": "The API service is currently unavailable."},
}


def raise_for_error(response):
    """Forming a response message for raising custom exception"""
    try:
        response_json = response.json()
    except Exception:
        response_json = {}

    error_code = response.status_code
    error_message = response_json.get("error_description") or response_json.get(
        "error", ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {})
    ).get("message", "An Unknown Error occurred,please try after some time.")

    message = f"HTTP-error-code: {error_code}, Error: {error_message}"

    # Raise GoogleQuotaExceededError if 403 error code returned due to QuotaExceeded
    response_error = json.dumps(response_json.get("error", error_message))
    if error_code == 403 and "quotaExceeded" in response_error:
        ex = GoogleQuotaExceededError
    elif error_code == 400 and "invalid_grant" in response_error:
        ex = GoogleInvalidGrant
        message = f"HTTP-error-code: {error_code}, Error: invalid_grant"
    else:
        ex = ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {}).get("raise_exception", GoogleError)
    raise ex(message) from None



#!/usr/bin/env python

from setuptools import find_packages, setup

setup(
    name="tap-google-search-console",
    version="1.1.0",
    description="Singer.io tap for extracting data from the Google Search Console API",
    author="jeff.huth@bytecode.io",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    py_modules=["tap_google_search_console"],
    install_requires=["backoff==2.2.1", "requests==2.31.0", "singer-python==6.0.0"],
    extras_require={
        "dev": [
            "ipdb",
            "pylint",
        ]
    },
    entry_points="""
          [console_scripts]
          tap-google-search-console=tap_google_search_console:main
      """,
    packages=find_packages(),
    package_data={"tap_google_search_console": ["schemas/*.json"]},
)



#!/usr/bin/env python3
from singer import get_logger, utils

from tap_google_search_console.client import GoogleClient
from tap_google_search_console.discover import discover
from tap_google_search_console.sync import sync

LOGGER = get_logger()

REQUIRED_CONFIG_KEYS = ["client_id", "client_secret", "refresh_token", "start_date", "user_agent"]


@utils.handle_top_exception(LOGGER)
def main():
    # Parse command line arguments
    parsed_args = utils.parse_args(REQUIRED_CONFIG_KEYS)

    # If discover flag was passed, run discovery mode and dump output to stdout
    with GoogleClient(
        parsed_args.config["client_id"],
        parsed_args.config["client_secret"],
        parsed_args.config["refresh_token"],
        parsed_args.config["site_urls"],
        parsed_args.config["user_agent"],
        parsed_args.config.get("request_timeout"),
    ) as client:
        if parsed_args.discover:
            catalog = discover(client)
            catalog.dump()
        else:
            catalog = parsed_args.catalog or discover(client)
            sync(client, parsed_args.config, parsed_args.state, catalog)


if __name__ == "__main__":
    main()



from typing import Dict, List
import hashlib
import json
import os
import re
from urllib.parse import quote

import singer

LOGGER = singer.get_logger()


def get_abs_path(path: str):
    """Returns absolute path for URL."""
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def encode_and_format_url(url: str, string_format: str) -> str:
    """Encodes the given site_url."""
    return string_format.format(quote(url, safe=""))


def convert(name):
    """Converts a CamelCased word to snake case."""
    regsub = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", regsub).lower()


# Convert keys in json array
def convert_array(arr: List):
    """Converts all the CamelCased dict Keys in a list to snake case Iterated
    through each object recursively and if the object is dict type then
    converts its key to snake case."""
    new_arr = []
    for element in arr:
        if isinstance(element, list):
            new_arr.append(convert_array(element))
        elif isinstance(element, dict):
            new_arr.append(convert_json(element))
        else:
            new_arr.append(element)
    return new_arr


# Convert keys in json
def convert_json(data_object: Dict):
    """Converts all the CamelCased Keys in a nested dictionary object to snake
        case."""
    out = {}
    for key in data_object:
        new_key = convert(key)
        if isinstance(data_object[key], dict):
            out[new_key] = convert_json(data_object[key])
        elif isinstance(data_object[key], list):
            out[new_key] = convert_array(data_object[key])
        else:
            out[new_key] = data_object[key]
    return out


def remove_keys_nodes(data_object: Dict, path: str):
    """Remove `keys` node, if exists."""
    idx = 0
    for record in list(data_object[path]):
        if record.get("keys", None):
            data_object[path][idx].pop("keys")
        idx = idx + 1
    return data_object


def hash_data(data):
    """Create MD5 hash key for data element Prepares the project id hash."""
    hash_id = hashlib.md5()
    hash_id.update(repr(data).encode("utf-8"))
    return hash_id.hexdigest()


def denest_key_fields(data_object: Dict, stream_name: str, path: str, dimensions_list: List):
    """Denest keys values list to dimension_list keys"""
    idx = 0
    for record in list(data_object[path]):
        for key in list(record.keys()):
            if isinstance(record[key], list) and key == "keys":
                dim_num = 0
                # Add dimensions_hash_key for performance_report_custom
                if stream_name == "performance_report_custom":
                    dims_md5 = str(hash_data(json.dumps(record[key], sort_keys=True)))
                    data_object[path][idx]["dimensions_hash_key"] = dims_md5
                for dimension in dimensions_list:
                    data_object[path][idx][dimension] = record[key][dim_num]
                    dim_num = dim_num + 1
        idx = idx + 1
    return data_object


def add_site_url_search_type(data_object: Dict, path: str, site: str, key_name: str = "site_url"):
    """Adds site_url or Search_type key, value to each object
        Default of key_name is site_url"""
    idx = 0
    for _ in data_object[path]:
        data_object[path][idx][key_name] = site
        idx = idx + 1
    return data_object


def transform_reports(data_object: Dict, stream_name: str, path: str, site: str, sub_type: str, dimensions_list: List):
    """de-nest keys array to dimension fields and add MD5 hash key for custom report"""
    denested_json = denest_key_fields(data_object, stream_name, path, dimensions_list)
    # remove keys array node
    keyless_json = remove_keys_nodes(denested_json, path)
    return add_site_url_search_type(add_site_url_search_type(keyless_json, path, site), path, sub_type,
                                    key_name="search_type")


def transform_json(data_object: Dict, stream_name: str, path: str = "", site: str = "",
                   sub_type: str = "", dimensions_list: List = None):
    """Run all transforms: convert camelCase to snake_case for field_name keys,
    and stream-specific transforms for sitemaps and performance_reports."""
    dimensions_list = dimensions_list or []
    converted_json = convert_json(data_object)
    if stream_name == "sitemaps":
        return add_site_url_search_type(converted_json, path, site)
    elif stream_name.startswith("performance_report"):
        return transform_reports(converted_json, stream_name, path, site, sub_type, dimensions_list)
    return converted_json

